/***************************************************************************
                          wadview.c  -  description
                             -------------------
    begin                : Wed Jul 5 2000
    copyright            : (C) 2000 by Simon E. Reichelt
    email                : s.reichelt@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include "support.h"
#include "texdat.h"
#include "callbacks.h"


typedef struct
{
	char f_name[1024];
	texrep_t *tex_entry;
} finfo_t;


texrep_t *tex_dat = NULL;

GtkWidget *con_text;

GtkWidget *pck_table1 = NULL;

int tex_cnt = 0;

int sel_index = -1;

extern gboolean do_log;


void _CreateButton(GtkWidget **ok_button, GtkWidget *dialog)
{
	*ok_button = gtk_button_new_with_label("OK");
	
	gtk_widget_set_usize(*ok_button, 50, -1);
	
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
	                   *ok_button, FALSE, FALSE, 0);
	
	gtk_signal_connect_object(GTK_OBJECT(*ok_button), "clicked",
	                          GTK_SIGNAL_FUNC(gtk_widget_destroy),
	                          (void*) dialog);
	
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
}


int WVError(char message[], ...)
{
	va_list parms;
	char error_str[1024];
	int char_cnt;
	GtkWidget *dialog;
	GtkWidget *ok_button;
	GtkWidget *label1;
	
	va_start(parms, message);
	
	char_cnt = vsprintf(error_str, message, parms);
	
	dialog = gtk_dialog_new();
	
	gtk_window_set_title(GTK_WINDOW(dialog),
	                     "WadVision Error");
	label1 = gtk_label_new(error_str);
	
	_CreateButton(&ok_button, dialog);	
	
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),
	                   label1, FALSE, FALSE, 15);
	
	gtk_widget_set_usize(dialog, 200, 100);
	
	gtk_widget_show(label1);
	gtk_widget_show(ok_button);
	gtk_widget_show(dialog);
	
	va_end(parms);
	
	return char_cnt;
}


GdkColor *_AllocTextColor(void)
{
	GdkColormap *clr_map;
	static GdkColor txt_clr;
	
	clr_map = gdk_colormap_get_system();
	
	txt_clr.red = 0 * 256;
	txt_clr.green = 21 * 256;
	txt_clr.blue = 182 * 256;
	
	gdk_colormap_alloc_color(clr_map, &txt_clr, FALSE, FALSE);
	
	return &txt_clr;
}	


int ConPrint(const char message[], ...)
{
	va_list parms;
	char msg_str[1024];
	int char_cnt, insert_p;
	GdkColor *txt_clr;
	FILE *log_file;
	char log_name[2048];
	
	va_start(parms, message);
	
	char_cnt = vsprintf(msg_str, message, parms);
	
	va_end(parms);
	
	txt_clr = _AllocTextColor();
	
	insert_p = gtk_text_get_point(GTK_TEXT(con_text));
	gtk_text_set_point(GTK_TEXT(con_text), insert_p);
	
	gtk_text_insert(GTK_TEXT(con_text), NULL, txt_clr,
	                NULL, msg_str, -1);
	
	if (do_log)
	{
		strcpy(log_name, getenv("HOME"));
		strcat(log_name, "/wadvision.log");
		
		log_file = fopen(log_name, "a");
		if (!log_file)
		{
			WVError("Couldn't Create Log File!");
			return char_cnt;
		}
		
		fputs(msg_str, log_file);
		fclose(log_file);
	}
	
	return char_cnt;
}


void PrepareRectGC(GdkGC *gc)
{
	GdkColormap *clr_map;
	GdkColor rect_clr;
	
	clr_map = gdk_colormap_get_system();
	
	rect_clr.red = 0;
	rect_clr.green = 0;
	rect_clr.blue = 65535;
	
	gdk_colormap_alloc_color(clr_map, &rect_clr, FALSE, FALSE);
	gdk_gc_set_foreground(gc, &rect_clr);
	
	gdk_gc_set_line_attributes(gc, 4, GDK_LINE_SOLID,
	                           GDK_CAP_PROJECTING,
	                           GDK_JOIN_MITER);
}


void SetSizeLabel(GtkWidget *widget, texrep_t *tex_entry)
{
	int width, height;
	char size_str[124];
	GtkWidget *s_label;
	
	GetRealSize(tex_entry, &width, &height);
	
	s_label = lookup_widget(widget, "label2");
	
	sprintf(size_str, "%d x %d", width, height);
	
	gtk_label_set_text(GTK_LABEL(s_label), size_str);
}	


void SelectTexture(texrep_t *tex_entry, GtkWidget *emitter, int *s_index)
{
	GdkGC *gc;
	
	if (tex_entry->index == *s_index)
		return;  //save time if tex was selected twice
	
	gc = gdk_gc_new(emitter->window);
	
	if (*s_index != -1)  //only access FILLED buffer
	{		
		gtk_widget_unmap(tex_dat[*s_index].d_area);
		gtk_widget_map(tex_dat[*s_index].d_area);
	}
	
	PrepareRectGC(gc);  //set rectangle color and line width
	
	gdk_draw_rectangle(tex_entry->d_area->window, gc, FALSE, 2, 2,
	                   tex_entry->width - 4, tex_entry->height - 4);
	
	SetSizeLabel(emitter, tex_entry);
	
	*s_index = tex_entry->index;  //update selection index
	
	if (wad_type == WAD3)
		if (!open_cmap)
		{
			gtk_widget_unmap(cmap_canvas);
			gtk_widget_map(cmap_canvas);
		}
}


int RedrawDetail(GtkWidget *emitter, GdkEvent *event, gpointer data)
{
	texrep_t *tex_entry;
	GdkGC *gc = gdk_gc_new(emitter->window);
	byte_t *pxm_dat;
	int width, height;
	
	tex_entry = (texrep_t*) data;
	
	if (tex_entry->zero_offs == NULL)  //check for full-size image
		pxm_dat = tex_entry->pxm_data;
	else
		pxm_dat = tex_entry->zero_offs;
		
	GetRealSize(tex_entry, &width, &height);
	
	gdk_draw_indexed_image(emitter->window, gc, 0, 0, width,
	                       height, GDK_RGB_DITHER_NORMAL,
	                       pxm_dat, width,
	                       tex_entry->clr_map);
	
	return TRUE;
}	

	
void OpenDetailDialog(GtkWidget *widget, texrep_t *tex_entry)
{
	GtkWidget *d_dialog;
	GtkWidget *d_area;
	GtkWidget *e_box;
	GtkWidget *close_b;
	int a_width, a_height;
	
	GetRealSize(tex_entry, &a_width, &a_height);
	
	if (a_width <= 64 && a_height <= 64)
		return;
		
	d_dialog = gtk_dialog_new();
	
	gtk_window_set_title(GTK_WINDOW(d_dialog),
	                     tex_entry->name);
	gtk_window_set_modal(GTK_WINDOW(d_dialog), TRUE);
	
	if (a_width <= 64)
		gtk_widget_set_usize(d_dialog, 140, -1);
	
	gtk_window_set_policy(GTK_WINDOW(d_dialog), FALSE,
	                      FALSE, FALSE);
	
	d_area = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(d_area),
	                      a_width, a_height);
	
	gtk_widget_set_events(d_area, GDK_EXPOSURE_MASK);	
		
	gtk_signal_connect(GTK_OBJECT(d_area), "expose_event",
	                   GTK_SIGNAL_FUNC(RedrawDetail),
	                   (gpointer) tex_entry);
	e_box = gtk_hbox_new(TRUE, 14);
	gtk_container_set_border_width(GTK_CONTAINER(e_box),
	                               14);
	gtk_box_pack_start(GTK_BOX(e_box), d_area, TRUE, FALSE, 0);
	
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(d_dialog)->vbox), e_box,
	                   TRUE, FALSE, 0);
	
	close_b = gtk_button_new_with_label("Close");
	
	gtk_signal_connect_object(GTK_OBJECT(close_b), "clicked",
	                          GTK_SIGNAL_FUNC(gtk_widget_destroy),
	                          (void*) d_dialog);
	gtk_widget_set_usize(close_b, 44, -1);
	
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(d_dialog)->action_area),
	                   close_b, TRUE, FALSE, 10);
	
	gtk_widget_show(d_area);
	gtk_widget_show(close_b);
	gtk_widget_show(e_box);
	gtk_widget_show(d_dialog);
}


void SetTrue(GtkWidget *emitter, gpointer data)
{
	finfo_t *f_info;
	
	f_info = (finfo_t*) data;
	
	gtk_widget_destroy(gtk_widget_get_toplevel(emitter));
	
	WriteSGIFile(f_info->f_name, f_info->tex_entry);
}


void WriteActionDialog(char f_name[], texrep_t *tex_entry)
{
	GtkWidget *dialog;
	GtkWidget *ok_button;
	GtkWidget *cancel_button;
	GtkWidget *label1;
	static finfo_t f_info;
	
	strcpy(f_info.f_name, f_name);
	f_info.tex_entry = tex_entry;
	
	dialog = gtk_dialog_new();
	
	gtk_window_set_title(GTK_WINDOW(dialog), "Info");
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	gtk_widget_set_usize(dialog, 200, 100);
	
	label1 = gtk_label_new("The requested file already exists.\n"
	                       "Should I overwrite it?");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label1,
	                   FALSE, FALSE, 15);
	
	ok_button = gtk_button_new_with_label("OK");
	gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
	                   GTK_SIGNAL_FUNC(SetTrue), &f_info);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
	                   ok_button, TRUE, FALSE, 0);
	gtk_widget_set_usize(ok_button, 50, -1);
	
	cancel_button = gtk_button_new_with_label("Cancel");
	gtk_signal_connect_object(GTK_OBJECT(cancel_button), "clicked",
	                          GTK_SIGNAL_FUNC(gtk_widget_destroy),
	                          (gpointer) dialog);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
	                   cancel_button, TRUE, FALSE, 0);
	gtk_widget_set_usize(cancel_button, 50, -1);
	
	gtk_widget_show(label1);
	gtk_widget_show(ok_button);
	gtk_widget_show(cancel_button);
	gtk_widget_show(dialog);
}
	

void PreWriteSGIFile(GtkWidget *emitter, gpointer data)
{
	GtkWidget *file_d;
	char f_name[1024];
	FILE *sgi_file;
	
	file_d = gtk_widget_get_toplevel(emitter);
	
	strcpy(f_name, gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_d)));
	
	gtk_widget_destroy(file_d);
	
	sgi_file = fopen(f_name, "r");
	if (sgi_file)
	{
		WriteActionDialog(f_name, (texrep_t*) data);
		return;
	}
		
	if (sgi_file)
		fclose(sgi_file);
	
	WriteSGIFile(f_name, (texrep_t*) data);
}


void ExportTexture(GtkWidget *emitter, gpointer data)
{
	GtkWidget *file_d;
	texrep_t *tex_entry = (texrep_t*) data;
	char f_name[256];
	
	strcpy(f_name, tex_entry->name);
		
	file_d = gtk_file_selection_new("Save As SGI");
	
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_d)->cancel_button),
	                          "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
	                          (gpointer) file_d);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_d)->ok_button),
	                   "clicked", GTK_SIGNAL_FUNC(PreWriteSGIFile),
	                   data);
	
	strcat(f_name, ".sgi");
	
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_d),
	                                f_name);
	
	gtk_widget_show(file_d);
}


void WriteOutput(GtkWidget *emitter, gpointer data)
{
	char out_dir[1500];
	char out_name[2048];
	char w_dir[4000];
	int i, max_tex;
	
	getcwd(w_dir, 4000);
	
	strcpy(out_dir, gtk_file_selection_get_filename(GTK_FILE_SELECTION(data)));
	
	gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(data)));
	
	if (out_dir[strlen(out_dir) - 1] != '/')
	{
		WVError("You Didn't Select A Directory!");
		return;
	}
	else if (chdir(out_dir) == -1)
	{
		WVError("Couldn't Access Directory!");
		return;
	}
	
	chdir(w_dir);
	
	max_tex = tex_cnt;
	if (wad_type == WAD2)
		max_tex--;
	
	for (i = 0; i < max_tex; i++)
	{
		strcpy(out_name, out_dir);
		strcat(out_name, tex_dat[i].name);
		strcat(out_name, ".sgi");
	
		if (WriteSGIFile(out_name, &tex_dat[i]) == -1)
			return;
	}
}


void ExportAllTextures(GtkWidget *emitter, gpointer data)
{
	GtkWidget *file_d;
	
	file_d = gtk_file_selection_new("Select Output Directory");
	
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_d)->ok_button),
	                   "clicked", GTK_SIGNAL_FUNC(WriteOutput),
	                   file_d);
	
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_d)->cancel_button),
	                          "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
	                          (gpointer) file_d);
	
	gtk_widget_show(file_d);
}


void OpenPopupMenu(GdkEvent *event, texrep_t *tex_entry)
{
	GtkWidget *menu1, *n_menu;
	GtkWidget *m_items[5], *n_item;
	int i;
	
	menu1 = gtk_menu_new();
	
	m_items[0] = gtk_menu_item_new_with_label("Full-Size View");
	m_items[1] = gtk_menu_item_new_with_label("Full Name");
	m_items[2] = gtk_menu_item_new();
	m_items[3] = gtk_menu_item_new_with_label("Export To SGI");
	m_items[4] = gtk_menu_item_new_with_label("Export All...");
	
	gtk_signal_connect(GTK_OBJECT(m_items[3]), "activate",
	                   GTK_SIGNAL_FUNC(ExportTexture),
	                   (gpointer) tex_entry);
	gtk_signal_connect(GTK_OBJECT(m_items[0]), "activate",
	                   GTK_SIGNAL_FUNC(OpenDetailDialog),
	                   (gpointer) tex_entry);
	gtk_signal_connect(GTK_OBJECT(m_items[4]), "activate",
	                   GTK_SIGNAL_FUNC(ExportAllTextures),
	                   NULL);
	
	n_menu = gtk_menu_new();
	n_item = gtk_menu_item_new_with_label(tex_entry->name);
	
	gtk_menu_append(GTK_MENU(n_menu), n_item);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(m_items[1]),
	                          n_menu);
	
	gtk_widget_show(n_item);
	
	for (i = 0; i < 5; i++)
	{
		gtk_menu_append(GTK_MENU(menu1), m_items[i]);			
		gtk_widget_show(m_items[i]);
	}
	
	if (!tex_entry->zero_offs)
		gtk_widget_set_sensitive(m_items[0], FALSE);	
		
	gtk_menu_popup(GTK_MENU(menu1), NULL, NULL, NULL, NULL,
	               event->button.button, event->button.time);
}


int MouseEvent(GtkWidget *emitter, GdkEvent *event, gpointer data)
{
	texrep_t *tex_entry;
	
	tex_entry = (texrep_t*) data;
	
	if (event->type == GDK_2BUTTON_PRESS &&
	    event->button.button == 1)
		OpenDetailDialog(emitter, tex_entry);
		
	if (event->type == GDK_BUTTON_PRESS &&
	    (event->button.button == 1 || event->button.button == 3))
	{		
		SelectTexture(tex_entry, emitter, &sel_index);
		
		if (event->button.button == 3)
			OpenPopupMenu(event, tex_entry);
	}
		
	return TRUE;
}


void SetGreyColor(GtkWidget *widget)
{
	GtkStyle *new_s;
	GdkColormap *clr_map;
	GdkColor bck_clr;
	GdkGC *gc;
	GtkWidget *main_w;
	
	main_w = lookup_widget(widget, "window1");
	
	clr_map = gdk_colormap_get_system();
	
	bck_clr.red = 35000;
	bck_clr.green = 35000;
	bck_clr.blue = 35000;
	
	gdk_color_alloc(clr_map, &bck_clr);
	
	gc = gdk_gc_new(main_w->window);
	gdk_gc_set_background(gc, &bck_clr);
	
	new_s = gtk_style_new();
	new_s->bg_gc[GTK_STATE_NORMAL] = gc;
	
	new_s->bg[GTK_STATE_NORMAL] = bck_clr;
	
	gtk_widget_set_style(widget, new_s);
}


GtkWidget *CreateParentTable(GtkWidget *widget, int tex_num)
{
	GtkWidget *v_port1;
	GtkWidget *p_table;
	int t_rows;
	
	v_port1 = lookup_widget(widget, "viewport1");
	p_table = lookup_widget(widget, "table1");
	
	gtk_widget_destroy(p_table);
	
	t_rows = tex_num / 4;
	
	if ((tex_num % 4) > 0)
		t_rows++;
		
	p_table = pck_table1 = gtk_table_new(t_rows, 4, TRUE);	
	
	gtk_table_set_row_spacings(GTK_TABLE(p_table), 6);
	gtk_table_set_col_spacings(GTK_TABLE(p_table), 8);
	
	gtk_container_add(GTK_CONTAINER(v_port1), p_table);
	
	SetGreyColor(v_port1);
	
	gtk_widget_show(p_table);
	
	return p_table;
}
	
	
void PopulateTexBrowser(texrep_t *tex_set, int tex_num, void *data)
{
	GtkWidget *p_table;
	GtkWidget *vbox1;
	GtkWidget *d_area1;
	GtkWidget *label1;
	GtkWidget *tx_socket;
	int i, max_tex;
	int x = 0, y = 0;
	char label_txt[20];
	
	p_table = CreateParentTable(GTK_WIDGET(data), tex_num);
	
	max_tex = tex_num;
	
	if (wad_type == WAD2)
		max_tex--;
	
	for (i = 0; i < max_tex; i++)
	{
		vbox1 = gtk_vbox_new(0, FALSE);
		d_area1 = gtk_drawing_area_new();
		gtk_drawing_area_size(GTK_DRAWING_AREA(d_area1), tex_set[i].width, 64);
		gtk_widget_set_events(d_area1, GDK_EXPOSURE_MASK |
		                               GDK_BUTTON_PRESS_MASK);
		
		gtk_signal_connect(GTK_OBJECT(d_area1), "expose_event",
		                   GTK_SIGNAL_FUNC(RedrawTex), &tex_set[i]);
		gtk_signal_connect(GTK_OBJECT(d_area1), "button_press_event",
		                   GTK_SIGNAL_FUNC(MouseEvent), &tex_set[i]);
		
		TruncateTextureName(label_txt, tex_set[i].name);
		
		label1 = gtk_label_new(label_txt);
		tex_set[i].d_area = d_area1;
		tx_socket = gtk_hbox_new(0, TRUE);
		
		gtk_box_pack_start(GTK_BOX(tx_socket), d_area1, TRUE, FALSE, 0);
		gtk_box_pack_start(GTK_BOX(vbox1), tx_socket, FALSE, FALSE, 0);
		gtk_box_pack_start(GTK_BOX(vbox1), label1, FALSE, FALSE, 0);
		
		gtk_widget_set_usize(vbox1, 72, -1);
		
		gtk_table_attach(GTK_TABLE(p_table), vbox1, x, x + 1, y, y + 1,
		                 0, 0, 0, 0);
		x++;
		
		if (x == 4)  //check for row end
		{
			x = 0;
			y++;
		}
		
		gtk_widget_show(label1);
		gtk_widget_show(tx_socket);
		gtk_widget_show(d_area1);
		gtk_widget_show(vbox1);
	}
	
	gtk_widget_unmap(pck_table1);
	gtk_widget_map(pck_table1);
}


void SetMainWinTitle(GtkWidget *widget, char f_name[])
{
	GtkWidget *m_win;
	int i, j = 511;
	char file_str[512];
	char title_str[512] = "WadVision - ";
	
	m_win = gtk_widget_get_toplevel(widget);
	
	file_str[j--] = '\0';
	
	for (i = strlen(f_name) - 1; f_name[i] != '/' && i >= 0; i--)
		file_str[j--] = f_name[i];  //extract file name
		
	strcat(title_str, &file_str[++j]);
	
	gtk_window_set_title(GTK_WINDOW(m_win), title_str);
}


void InitWadLoad(GtkWidget *emitter, gpointer data)
{
	char f_name[256];
	GtkWidget *f_dialog;
	
	f_dialog = gtk_widget_get_toplevel(emitter);
	
	strcpy(f_name, gtk_file_selection_get_filename(
	       GTK_FILE_SELECTION(f_dialog)));
	
	if (f_name[strlen(f_name) - 1] == '/')
		return;
	
	gtk_widget_destroy(f_dialog);
	
	LoadWadFile(GTK_WIDGET(data), f_name);
}